/*
 *  This file is part of BGX, the Bayesian Gene eXpression program.
 *  Copyright (c) 2003-2004  Graeme Ambler <graeme@ambler.me.uk>
 *                2004       Anne-Mette Hein <a.hein@imperial.ac.uk>
 *                2005-2008  Ernest Turro <ernest.turro@ic.ac.uk>
 *
 *  BGX is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License, version 2, as
 *  published by the Free Software Foundation.
 *
 *  BGX is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "rand.hh"
#include "rundir.hh"
#include "qnorm.h"
#include "sokal.hh"
#include "bgx_updates.hh"
#include <string>
#include <cstring>
#include <iostream>
#include <fstream>
#include <valarray>
#include <vector>
#include <time.h>
#ifdef _OPENMP
  #include <omp.h>
  #define OMP_GET_MAX_THREADS omp_get_max_threads()
#else
  #define OMP_GET_MAX_THREADS 1
#endif // _OPENMP
#include "bgx.hh"
#include "bgx_frontend.hh"

#ifdef USING_R
  extern "C" {
  #include <R.h> // for flushing console, allowing user interrupts, and printing to console
  #if ( defined(HAVE_AQUA) || defined(WIN32) )
    #define FLUSH {R_FlushConsole(); R_ProcessEvents();}
  #else
    #define FLUSH R_FlushConsole();
  #endif
  #ifdef WIN32 // Win32 GUI doesn't handle \r properly. Use \b's.
    #define CARRIAGERETURN "\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b\b"
  #else
    #define CARRIAGERETURN "\r"
  #endif
  #define PRINTF(...) Rprintf((char*) __VA_ARGS__)
  #define EPRINTF(...) REprintf((char*) __VA_ARGS__)
  #define R_NO_REMAP
  #include <Rinternals.h>
  #include <R_ext/Rdynload.h>
  #include "writePMs.hh"
  #include "addByColumn.hh"
  }
  /* Register routines, allocate resources. */
  static const R_CMethodDef cMethods[] = {
    {"writePMs", (DL_FUNC)&writePMs,6},
    {"addByColumn", (DL_FUNC)&addByColumn,5},
    {"sokal", (DL_FUNC)&sokal,5},
    {"mmbgxFromR", (DL_FUNC)&mmbgxFromR,2},
    {NULL, NULL, 0}
  };
  void R_init_bgx(DllInfo *info) {
    R_registerRoutines(info, cMethods,NULL,NULL,NULL);
  }
  void R_unload_bgx(DllInfo *info) { } 
#else
  #define FLUSH fflush(stdout);
  #define PRINTF(...) printf(__VA_ARGS__)
  #define EPRINTF(...) fprintf(stderr, __VA_ARGS__)
  #define CARRIAGERETURN "\r"
#endif

using namespace std;

typedef valarray<double> array;
typedef valarray<valarray<double> > array2d;

void stringcpy(char* p, const string& s) {
  s.copy(p,string::npos);
  p[s.length()] = 0; // Add C-style string terminator
}

/* inline double TN_quantile(double p, double mu, double sigma) {
  return mu-sigma*qnorm(pnorm(mu/sigma)*(1-p));
} */


// The following are global so that we can delete them in freeBGXMemory if there is a user interrupt in R
static double *xave, *muave, *sacc, *muacc, *sigmaacc;
static ofstream *sigma_, *b_;
static fstream *mu_;

// Metropolis/Gibbs objects
static  S_T *AccS;
static  RWM<S_T,array2d> *S;
static  Mu_T *AccMu;
static  RWM<Mu_T,array2d> *Mu;
static  Sigma_T *AccSigma;
static  RWM<Sigma_T,array2d> *Sigma;
static B_T *B;

void bgx(double* pm, int* samples, int* conditions, 
	 int* probes, int* genes, int *numberCategories, int *samplesets, vector<int> * pm2ps, vector<int>* ps2pm,
   double *gamma_, double *delta_ ,
   int* categories,  
    int* iter, int* burnin, double* s_jmp,
	 double* mu_jmp, double* sigma_jmp, bool* adaptive, int* batch_size, double* optimalAR, 
	 int* subsample, char** dirname, char **basepath, int* seed, char **sampleNames, vector<string> filesToCopy) {

  PRINTF("Maximum number of threads: %d\n", OMP_GET_MAX_THREADS);
  // SETUP. declaration/initialisation of some variables up here for common code structure with parallel BGX
  string run_dir;

  array2d PM(*samples), s(*samples), gamma(*samples), delta(*samples),
    mu(*conditions), sigma(*conditions), 
    s_jmps(*samples), mu_jmps(*conditions), sigma_jmps(*conditions);

  for(int j=0; j < *samples; ++j) {
    PM[j].resize(*probes);
    s[j].resize(*probes);
    s_jmps[j].resize(*probes);
    gamma[j].resize(*numberCategories);
    delta[j].resize(*numberCategories);
    for(int i=0; i < *probes; i++) {
      s_jmps[j][i]=*s_jmp;
    }
  }
  for(int j=0; j < *conditions; ++j) {
    mu[j].resize(*genes);
    mu_jmps[j].resize(*genes);
    sigma[j].resize(*genes);
    sigma_jmps[j].resize(*genes);
    for(int i=0; i<*genes; ++i){
      mu_jmps[j][i]=*mu_jmp;
      sigma_jmps[j][i]=*sigma_jmp;
    }
  }

  array a(*conditions), b(*conditions);
  
  ofstream muave_, xave_, sacc_, muacc_, sigmaacc_,
    s_jmps_, mu_jmps_, sigma_jmps_;

  // Initialise "averages"
  muave = new double[*genes**conditions];
  for(int i=0; i<*genes**conditions; ++i) muave[i]=0;

  xave = new double[*probes**samples];
  for(int i=0; i<*probes**samples; ++i) xave[i]=0;

  // END SETUP

  Random *rand[OMP_GET_MAX_THREADS];
  for(int i=0; i < OMP_GET_MAX_THREADS; i++) {
      rand[i] = new Random(*seed+i);
  }

  // Create output file and directory
  char *tmpstr = new char[strlen(*basepath)+4];
  strcpy(tmpstr, *basepath); // Prepend basepath
  strcat(tmpstr,"/run");
  run_dir = rundir(tmpstr);
  delete [] tmpstr;

  // In R, dirname is set to the ouput directory
  stringcpy(*dirname,run_dir);

 // Copy files to run_dir
  for(int i =0 ; i < (signed)filesToCopy.size(); i++) {
    ifstream in(filesToCopy[i].c_str());
    unsigned long int lastslash = filesToCopy[i].find_last_of('/');
    if(lastslash != string::npos) filesToCopy[i] = filesToCopy[i].substr(lastslash+1);;
    ofstream out((run_dir + '/' + filesToCopy[i]).c_str());
    out << in.rdbuf();
  }

  // Write out a summary of the call
  string filename = run_dir+"/summary.txt";
  ofstream summary(filename.c_str());
  summary << "The program was called with the following parameter values:" << endl << endl
	  << "Number of samples:\t" << *samples << endl
	  << "Number of conditions:\t" << *conditions << endl
	  << "Number of probes:\t" << *probes << endl
	  << "Number of probesets:\t" << *genes << endl 
    << "Number of categories:\t" << *numberCategories << endl
	  << "Subsampling interval:\t" << *subsample << endl
	  << "Number of burn-in sweeps:\t" << *burnin << endl
	  << "Number of sampling sweeps:\t" << *iter << endl
	  << "Initial spread of jumps in S:\t" << *s_jmp << endl
	  << "Initial spread of jumps in mu:\t" << *mu_jmp << endl
	  << "Initial spread of jumps in sigma:\t" << *sigma_jmp << endl
	  << "Adaptive MCMC:\t" << *adaptive << endl
	  << "Batch size:\t" << *batch_size << endl
	  << "Optimal acceptance ratio:\t" << *optimalAR << endl
	  << "Name of output directory:\t" << run_dir << endl
	  << "Seed for random number generator:\t" << *seed << endl
    << "CEL files:\t";
    for(int s =0; s < *samples-1; s++) summary << sampleNames[s] << ", ";
    summary << sampleNames[*samples-1] << endl;

  // Open output files
  filename=run_dir+"/muave";
  muave_.open(filename.c_str());
  mu_ = new fstream[*conditions];
  string num;
  for(int j=0; j<*conditions; ++j){
    int_to_string(j+1,num);
    filename=run_dir+"/mu."+num;
    // Open for in+out because we want to read these again later to calculate MCSE
    mu_[j].open(filename.c_str(), fstream::in | fstream::out | fstream::trunc);
  }
    
  filename=run_dir+"/xave";
  xave_.open(filename.c_str());

  s_jmps_.open((run_dir + "/s_jmps").c_str());
  mu_jmps_.open((run_dir + "/mu_jmps").c_str());
  sigma_jmps_.open((run_dir + "/sigma_jmps").c_str());
  
  sigma_ = new ofstream[*conditions];
  for(int j=0; j<*conditions; ++j){
    int_to_string(j+1,num);
    filename=run_dir+"/sigma2."+num;
    sigma_[j].open(filename.c_str());
  }

  b_ = new ofstream((run_dir+"/b2").c_str());

  // Acceptance probabilities
  sacc = new double[*probes**samples];
  muacc = new double[*genes**conditions];
  sigmaacc = new double[*genes**conditions];
  for(int i=0; i<*probes**samples; ++i)
    sacc[i]=0;
  for(int i=0; i<*genes**conditions; ++i)
    muacc[i]=sigmaacc[i]=0;

  // Open output files
  filename=run_dir+"/sacc";
  sacc_.open(filename.c_str());
  filename=run_dir+"/muacc";
  muacc_.open(filename.c_str());
  filename=run_dir+"/sigmaacc";
  sigmaacc_.open(filename.c_str());

  // Read data into C++-style arrays for ease of manipulation
  for(int j=0; j<*samples; ++j){
    for(int i=0; i<*probes; ++i) PM[j][i]=*(pm+i+*probes*j);
    for(int i=0; i < *numberCategories; i++){
      gamma[j][i]=*(gamma_+i+*numberCategories*j);
      delta[j][i]=*(delta_+i+*numberCategories*j);
    }
  }

  // Define parameters and pick some vaguely sensible values for 
  // the initial values.
  bool badPRNG=false;
  for(int j=0; j < *samples; j++) {
    for(int i=0; i < *probes; i++) {
      int count=0;
      while((s[j][i] = PM[j][i] - exp(rand[0]->Normal(gamma[j][categories[i]], sqrt(delta[j][categories[i]])))) <= 1) {
        count++;
        if(count==1000000) {
          // Lost cause, just set s to 1
          s[j][i] = 1;
          break;
        }
      }

      if(isnan(s[j][i])) {
        EPRINTF("Encountered NANs. Something wrong with PRNG.\n");
        badPRNG=true; // exit later after more allocations, so that freeBGXMemory is OK
      }
    }
  }

  double tempS, tempSS, templvS, templvSS;
  int offset=0;
  for(int c=0; c < *conditions; c++) {
    if(c>0) offset += samplesets[c-1];
    vector<int> singleProbePS;
    double averageSigma = 0;
    for(int g=0; g < *genes; g++) {
      tempS = tempSS = 0;
      for(int sam=offset; sam < offset+samplesets[c]; ++sam){
        for(unsigned int p=0; p < ps2pm[g].size(); p++) {
          tempS += log(s[sam][ps2pm[g][p]]/pm2ps[ps2pm[g][p]].size());
          tempSS += pow(log(s[sam][ps2pm[g][p]]/pm2ps[ps2pm[g][p]].size()), 2);
        }
      }
      double n = samplesets[c] *  ps2pm[g].size();
      // in the rare case where S_{jcr}/|G(j)| is consistently less than 1, set mu to minimum bound, zero
      mu[c][g] = tempS < 0 ? 0 : tempS/n;
      sigma[c][g] = 1.0/(n-1.0)*tempSS - n/(n-1.0)*mu[c][g]*mu[c][g];
      if(n==1 || sigma[c][g]==0) {
        /* this can happen if the PS contains only one probe (potentially self-cross-hybridising)
         * and there is only one sample per condition */
        singleProbePS.push_back(g); // save PS index to replace later with averageSigma
      } else {
        averageSigma += sigma[c][g];
      }
    }

    averageSigma /= (*genes - singleProbePS.size());
    for(unsigned int i=0; i < singleProbePS.size(); i++) {
      sigma[c][singleProbePS[i]] = averageSigma;
    }

    templvS = templvSS = 0;
    for(int g=0; g < *genes; g++) {
      templvS += log(sigma[c][g]);
      templvSS += log(sigma[c][g])*log(sigma[c][g]);
    }

    a[c] = templvS / double(*genes);
    b[c] = 1.0/(*genes-1.0)*templvSS - *genes/(*genes-1.0)*a[c]*a[c];
  }

  filename=run_dir+"/empBayesEst";
  ofstream empBayesEst_(filename.c_str());
  empBayesEst_ << "Empirical Bayes like estimates of a for each condition:" 
               << endl << endl;
  empBayesEst_ << '\t' << "condition" <<  '\t' << "a"  <<  endl;
  for(int j=0; j<*conditions; j++){
    empBayesEst_ << '\t'  << j << '\t' <<  a[j] << endl;
  }
  empBayesEst_.close();

  // Initialise the update objects. For random walk Metropolis (RWM) steps, this 
  // involves creating an acceptance probability functor (e.g. S_T AccS(...)) 
  // and a RWM object (e.g. RWM<S_T,array2d> S(...)). For Gibbs moves we just
  // create a single update object (e.g. B_T B(...)).
  // These are all defined in bgx_updates.hh.
  AccS = new S_T(PM,mu,sigma,gamma,delta,samplesets,categories,pm2ps,ps2pm );
  S = new RWM<S_T,array2d>(s,*AccS,s_jmps,*batch_size,*optimalAR,*s_jmp,rand);

  AccMu = new Mu_T(s,sigma,samplesets, pm2ps, ps2pm);
  Mu = new RWM<Mu_T,array2d>(mu,*AccMu,mu_jmps,*batch_size,*optimalAR,*mu_jmp,rand);

  AccSigma = new Sigma_T(s,mu,a,b,samplesets, pm2ps, ps2pm);
  Sigma = new RWM<Sigma_T,array2d>(sigma,*AccSigma,sigma_jmps,*batch_size,*optimalAR,*sigma_jmp,rand);

  B = new B_T(b,sigma,a,1.2,0.3,rand[0]);

  if(badPRNG) {
#ifdef USING_R
    return;
#else
    exit(1);
#endif
  }

  PRINTF("Starting MCMC simulation...\n");

  time_t start_time, end_time;
  time(&start_time);

//  clock_t cl;

  // Run the Markov chain
  for(int sweep=0; sweep<*burnin+*iter; ++sweep){
    if(sweep<*burnin){  
      PRINTF("%sBurn-in sweep %d of %d", CARRIAGERETURN, sweep , *burnin);
    } else {
      PRINTF("%sSampling sweep %d of %d\r", CARRIAGERETURN, sweep-*burnin, *iter);
    }
    if(sweep%16 == 0) FLUSH
#ifdef USING_R
    R_CheckUserInterrupt();
#endif

    // Output post burn-in subsample
    if(sweep >= *burnin) {
      if(!((sweep-*burnin)%*subsample)){

        #pragma omp parallel
        {
          // Update means of mu's 
          for(int c=0; c<*conditions; ++c){
            #pragma omp for schedule(static) nowait
            for(int g=0; g<*genes; ++g){
              muave[g+*genes*c] += mu[c][g];
            }
          }
          // Update means of log(S)'s
          for(int sample=0; sample<*samples; ++sample){
            #pragma omp for schedule(static) nowait
            for(int p=0; p<*probes; ++p){
              xave[p+*probes*sample] += log(s[sample][p]);
            }
          }

          #pragma omp sections nowait
          {
            #pragma omp section
            {
              // Output mu trace
              for(int c=0; c<*conditions; ++c){
                for(int g=0; g<*genes; ++g) {
                  mu_[c] << mu[c][g] << '\t';
                }
                mu_[c] << endl;
              }
            }

            #pragma omp section
            {
              // Output sigma's and b's
              for(int c=0; c<*conditions; ++c){
                for(int g=0; g<*genes; ++g) {
                  sigma_[c] << sigma[c][g] << '\t';
                }
                sigma_[c] << endl;
                *b_ << b[c] << '\t';
              }
              *b_ << endl;
            }
          }
        }
      }
    }

//    cl = clock();
    S->Update();
    Mu->Update();
    Sigma->Update();
    B->Update();
 //      cout << endl << (clock() - cl)/(double)CLOCKS_PER_SEC << endl;

    if(sweep<*burnin){  
      if(*adaptive && (sweep+1)%*batch_size==0){
        S->Update_jmps();
        Mu->Update_jmps();
        Sigma->Update_jmps();
      }
    }
  
    if(sweep==*burnin-1) {
      PRINTF("%s%d burn-in sweeps completed.\n", CARRIAGERETURN, *burnin);
      S->Reset();
      Mu->Reset();
      Sigma->Reset();
    }
  }
  // End of simulation

  for(int i=0; i < OMP_GET_MAX_THREADS; i++) {
      delete rand[i];
  }

  time(&end_time);
  int duration = (int)difftime(end_time,start_time);

  PRINTF("%s%d sampling sweeps completed.\n", CARRIAGERETURN, *iter);
  PRINTF("MCMC duration (%d thread(s)): %dh %dm %ds\n", OMP_GET_MAX_THREADS, duration/3600, duration%3600/60, duration%60);

  summary << "MCMC duration (" <<  OMP_GET_MAX_THREADS << " thread(s)):\t" <<  duration/3600 << "h " << duration%3600/60 << "m " << duration%60 << "s\n";
  summary.close();


  for(int c=0; c<*conditions; ++c){
    for(int g=0; g<*genes; ++g){
      muave_ << muave[g+*genes*c]**subsample/(*iter) << '\t';
    }
    muave_ << endl;
  }
  muave_.close();

  for(int sample=0; sample<*samples; ++sample){
    for(int p=0; p<*probes; ++p){
      xave_ << xave[p+*probes*sample]**subsample/(*iter) << '\t';
    }
    xave_ << endl;
  }
  xave_.close();
  for(int sample=0; sample<*samples; ++sample){
    for(int p=0; p<*probes; ++p) {
      s_jmps_ << s_jmps[sample][p] << "\t";
    }
    s_jmps_ << endl;
  }
  s_jmps_.close();

  for(int c=0; c<*conditions; ++c) {
    for(int g=0;g<*genes;g++){
      mu_jmps_ << mu_jmps[c][g] << "\t";
      sigma_jmps_ << sigma_jmps[c][g] << "\t";
    }
    mu_jmps_ << endl;
    sigma_jmps_ << endl;
  }

  mu_jmps_.close();
  sigma_jmps_.close();

  // Calculate acceptance probabilities
  S->pAccept(*iter,sacc);
  Mu->pAccept(*iter,muacc);
  Sigma->pAccept(*iter,sigmaacc);
  for(int c=0; c<*conditions; ++c){
    for(int g=0; g<*genes; ++g){
      muacc_ << muacc[g+*genes*c] << '\t';
      sigmaacc_ << sigmaacc[g+*genes*c] << '\t';
    }
    muacc_ << endl;
    sigmaacc_ << endl;
  }

  for(int sample=0; sample<*samples; ++sample){
    for(int p=0; p<*probes; ++p){
      sacc_ << sacc[p+*probes*sample] << '\t';
    }
    sacc_ << endl;
  }
  sacc_.close();
  muacc_.close();
  sigmaacc_.close();

  // IACT and MCSE
  ofstream mumcse_((run_dir + "/mumcse").c_str());
  ofstream muiact_((run_dir + "/muiact").c_str());
  double * mumcse = new double[*genes * *iter/(*subsample)];
  for(int c=0; c < *conditions; c++) {
    mu_[c].seekg(0, ios::beg);
    
    for(int i=0; i < *iter/(*subsample); i++) {
      for(int g=0; g < *genes; g++) mu_[c] >> mumcse[g**iter/(*subsample) + i];
    }

    for(int g=0; g < *genes; g++) {
      double var=0;
      double tau=0;
      int m=0;
      int len=*iter/(*subsample);
      if(sokal(&len,&mumcse[g**iter/(*subsample)],&var,&tau,&m)!=0) {
        mumcse_ << *iter/(*subsample) << " ";
        muiact_ << "NA" << " ";
      } else {
        mumcse_ << sqrt(tau*var**subsample/(*iter)) << " ";
        muiact_ << tau << " ";
      }
    }
    mumcse_ << endl;
    muiact_ << endl;
  }
  mumcse_.close();
  muiact_.close();
  for(int c=0; c<*conditions; ++c) mu_[c].close();
  delete [] mumcse;

  PRINTF("CEL files analysed: \t");
  for(int s =0; s < *samples-1; s++) PRINTF("%s, ", sampleNames[s]);
  PRINTF("%s\n",sampleNames[*samples-1]);
  PRINTF("Results are in %s\n", *dirname);
  PRINTF("Run finished.\n");

#ifndef USING_R // If using R, let R call this function, otherwise call here
  freeBGXMemory();
#endif
}

void freeBGXMemory() {
  delete [] mu_;
  delete [] muave;
  delete AccS;
  delete S;
  delete AccMu;
  delete Mu;
  delete AccSigma;
  delete Sigma;
  delete B;

  delete [] sigma_;
  delete [] xave;

  delete [] muacc;
  delete [] sigmaacc;
  delete [] sacc;
}
